module fir_onboard (clkin, key_begin_n, dout, done, key_ini_n, key_rst_n, input_rdy, out_gather);
parameter tapsize = 1024;
parameter idata_width = 12;
parameter icoeffi_width = 24;
parameter sim_en = 0;

parameter coeffi_addr_width = $clog2(tapsize);
localparam dout_width = 54;

input clkin;
input key_begin_n;
input key_rst_n;
input key_ini_n;
output [dout_width - 1 : 0] dout;
output reg [dout_width - 1 : 0] out_gather;
output done;
output input_rdy;

reg signed [icoeffi_width - 1 : 0] coeffi [0 : tapsize - 1];
initial begin
    if (sim_en) begin
        $readmemb("../../project/src/data/icoeffi.dat",coeffi);
    end
    else begin
        $readmemb("icoeffi.dat",coeffi);
    end
end

reg signed [idata_width - 1 : 0] data [0 : 2048 - 1];
initial begin
    if (sim_en) begin
        $readmemb("../../project/src/data/idata.dat",data);
    end
    else begin
        $readmemb("idata.dat",data);
    end
end

reg signed [54 - 1 : 0] odata [0 : 2048 - 1];

wire [dout_width - 1 : 0]dout;
wire onboard_begin;
wire key_begin;
wire key_ini;
wire ini;
wire clk;
wire rst_n;
wire key_rst;
wire rst;

assign key_begin = ~key_begin_n;
assign key_ini = ~key_ini_n;
assign key_rst = ~key_rst_n;
assign rst_n = ~rst;
//---------I/O PORT of basic_FIR_top---------
reg wr_en_data;
//wire wr_en_coeffi;cnt_wr_coeffi_en
reg [icoeffi_width - 1 : 0] din_coeffi;
reg [idata_width - 1 : 0] din_data;
wire done;
wire input_rdy;

//------------intenal signal-------------
reg cnt_wr_coeffi_en = 0;
reg [coeffi_addr_width - 1 : 0] cnt_wr_coeffi = 0;
reg re_coeffi_en = 0;
reg [coeffi_addr_width - 1 : 0] cnt_re_coeffi = 0;

reg [coeffi_addr_width - 1 : 0] cnt_ini;
reg cnt_ini_en;

reg onboard_begin_flip1 = 0;
reg onboard_begin_flip2 = 0;
reg onboard_begin_flag = 0;
reg wr_coeffi_flag = 0;
reg cnt_wr_coeffi_en_flip1 = 0;
reg cnt_wr_coeffi_en_flip2 = 0;
reg cnt_wr_coeffi_en_flag = 0;
reg wr_data_en = 0;
reg [11 : 0] re_data_addr = 0;//read data memory address
reg [11 : 0] sto_odata_addr = 0;//storage odata address
reg [10 : 0] output_odata_addr = 0;//output odata address
reg [9 : 0] wr_en_delay;
reg output_odata_en;

//reg wre_data;

reg done_flip1;
reg done_flip2;
reg done_flag;
//--------------------test_start  onboard_begin_flag------------------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        onboard_begin_flip1 <= 0;
        onboard_begin_flip2 <= 0;
    end
    else begin
        onboard_begin_flip1 <= onboard_begin;
        onboard_begin_flip2 <= onboard_begin_flip1;
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        onboard_begin_flag <= 0;
    end
    else begin
        if (onboard_begin_flip1 && !onboard_begin_flip2) begin
            onboard_begin_flag <= 1;
        end
        else begin
            onboard_begin_flag <= 0;
        end
    end
end

//-----------input data to fir----------
//cnt_wr_coeffi_en
always @(posedge clk) begin
    if (!rst_n || ini) begin
        cnt_wr_coeffi_en_flip1 <= 0;
        cnt_wr_coeffi_en_flip2 <= 0;
    end
    else begin
        cnt_wr_coeffi_en_flip1 <= cnt_wr_coeffi_en;
        cnt_wr_coeffi_en_flip2 <= cnt_wr_coeffi_en_flip1;
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        cnt_wr_coeffi_en_flag <= 0;
    end
    else begin
        if (!cnt_wr_coeffi_en_flip1 && cnt_wr_coeffi_en_flip2) begin
            cnt_wr_coeffi_en_flag <= 1;
        end
        else begin
            cnt_wr_coeffi_en_flag <= 0;
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        wr_data_en <= 0;
    end
    else begin
        if (cnt_wr_coeffi_en_flag) begin
            wr_data_en <= 1;
        end
        else begin
            if (re_data_addr == 12'b1000_0000_0000) begin
                wr_data_en <= 0;
            end
            else begin
                wr_data_en <= wr_data_en;
            end
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        re_data_addr <= 0;
    end
    else begin
        if (wr_en_delay == 10'd300) begin
            re_data_addr <= re_data_addr + 1;
        end
        else begin
            re_data_addr <= re_data_addr;
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        wr_en_delay <= 0;
    end
    else begin
        if (wr_data_en) begin
            if (wr_en_delay == 10'd300) begin
                wr_en_delay <= 0;
            end
            else begin
                wr_en_delay <= wr_en_delay + 1;
            end
        end
        else begin
            wr_en_delay <= 0;
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        wr_en_data <= 0;
    end
    else begin
        if ((wr_en_delay == 10'd300) || (wr_en_delay == 10'd299)) begin
            wr_en_data <= 1;
        end
        else begin
            wr_en_data <= 0;
        end
    end
end

//----------------
//always @(posedge clk) begin
//    if (!rst_n || ini) begin
//        wre_data <= 0;
//    end
//    else begin
//        if (wr_en_delay == 9'd280) begin
//            wre_data <= 1;
 //       end
//        else begin
 //           wre_data <= 0;
//        end
//    end
//end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        din_data <= 0;
    end
    else begin
        din_data <= data[re_data_addr];
    end
end
//---------------wr_coeffi_flag write coeffi to fir----------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        wr_coeffi_flag <= 0;
    end
    else begin
        if (onboard_begin_flag) begin
            wr_coeffi_flag <= 1;
        end
        else begin
            wr_coeffi_flag <= 0;
        end
    end
end
//---------------------read coeffi ------------------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        re_coeffi_en <= 0;
    end
    else begin
        if (onboard_begin_flag) begin
            re_coeffi_en <= 1;
        end
        else begin
            if (cnt_re_coeffi == (tapsize - 1)) begin
                re_coeffi_en <= 0;
            end
            else begin
                re_coeffi_en <= re_coeffi_en;
            end
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        cnt_re_coeffi <= 0;
    end
    else begin
        if (re_coeffi_en) begin
            cnt_re_coeffi <= cnt_re_coeffi + 1;
        end
        else begin
            cnt_re_coeffi <= 0;
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        din_coeffi <= 0;
    end
    else begin
        if (re_coeffi_en) begin
            din_coeffi <= coeffi[cnt_re_coeffi];
        end
        else begin
            din_coeffi <= 0;
        end
    end
end

//---------------------- cnt_wr_coeffi_en -----------------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        cnt_wr_coeffi_en <= 0;
    end
    else begin
        if (wr_coeffi_flag) begin
            cnt_wr_coeffi_en <= 1;
        end
        else begin
            if (cnt_wr_coeffi == (tapsize - 1)) begin
                cnt_wr_coeffi_en <= 0;
            end
            else begin
                cnt_wr_coeffi_en <= cnt_wr_coeffi_en;
            end
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        cnt_wr_coeffi <= 0;
    end
    else begin
        if (cnt_wr_coeffi_en) begin
            cnt_wr_coeffi <= cnt_wr_coeffi + 1;
        end
        else begin
            cnt_wr_coeffi <= 0;
        end
    end
end

//-------------------initial-------------------
always @(posedge clk) begin
    if (!rst_n) begin
        cnt_ini_en <= 0;
    end
    else begin
        if (ini) begin
            cnt_ini_en <= 1;
        end
        else begin
            if (cnt_ini == tapsize - 1) begin
                cnt_ini_en <= 0;
            end
            else begin
                cnt_ini_en <= cnt_ini_en;
            end
        end
    end
end

always @(posedge clk) begin
    if (!rst_n) begin
        cnt_ini <= 0;
    end
    else begin
        if (cnt_ini_en) begin
            cnt_ini <= cnt_ini + 1;
        end
        else begin
            cnt_ini <= 0;
        end
    end
end


//-----------------out_put_gather--------------------
//---------done_flage-------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        done_flip1 <= 0;
        done_flip2 <= 0;
    end
    else begin
        done_flip1 <= done;
        done_flip2 <= done_flip1;
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        done_flag <= 0;
    end
    else begin
        if (done_flip1 && !done_flip2) begin
            done_flag <= 1;
        end
        else begin
            done_flag <= 0;
        end
    end
end

//--------------storage dout-------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        sto_odata_addr <= 0;
    end
    else begin
        if (done_flag) begin
            if (sto_odata_addr == 12'b0111_1111_1111) begin
                sto_odata_addr <= 12'b0111_1111_1111;
            end
            else begin
                sto_odata_addr <= sto_odata_addr + 1;
            end
        end
        else begin
            sto_odata_addr <= sto_odata_addr;
        end
    end
end

always @(posedge clk) begin
    if (done_flag) begin
        odata[sto_odata_addr] <= dout;
    end
end

//----------output odata--------------
always @(posedge clk) begin
    if (!rst_n || ini) begin
        output_odata_en <= 0;
    end
    else begin
        if ((sto_odata_addr == 12'b0111_1111_1111) && done_flag) begin
            output_odata_en <= 1;
        end
        else begin
            if (output_odata_addr == 11'b111_1111_1111) begin
                output_odata_en <= 0;
            end
            else begin
                output_odata_en <= output_odata_en;
            end
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        output_odata_addr <= 0;
    end
    else begin
        if (output_odata_en) begin
            if (output_odata_addr == 11'b111_1111_1111) begin
                output_odata_addr <= 0;
            end
            else begin
                output_odata_addr <= output_odata_addr + 1;
            end
        end
        else begin
            output_odata_addr <= 0;
        end
    end
end

always @(posedge clk) begin
    if (!rst_n || ini) begin
        out_gather <= 0;
    end
    else begin
        if (output_odata_en) begin
            out_gather <= odata[output_odata_addr];
        end
        else begin
            out_gather <= 0;
        end
    end
end



deUstb UUT_begin_deUstb(
    .out(onboard_begin),
    .in(key_begin),
    .clk(clk)
);

deUstb UUT_ini_deUstb(
    .out(ini),
    .in(key_ini),
    .clk(clk)
);

deUstb UUT_rst_deUstb(
    .out(rst),
    .in(key_rst),
    .clk(clk)
);

GW_PLL PLL_UUT(
    .clkout(clk), 
    .clkin(clkin)
);

Basic_FIR_Filter_Top basic_FIR_UUT(
    .clk(clk),
    .rst_n(rst_n),
    .ini(cnt_ini_en),
    .wr_en_data(wr_en_data),
    .wr_en_coeffi(cnt_wr_coeffi_en),
    .din_coeffi(din_coeffi),
    .din_data(din_data),
    .dout(dout),
    .done(done),
    .input_rdy(input_rdy)
    );

endmodule